/* ------------------------------------------------------------------------- */
/* tftp.c                                                                    */
/*                                                                           */
/* A simple tftp client for busybox.                                         */
/* Tries to follow RFC1350.                                                  */
/* Only "octet" mode and 512-byte data blocks are supported.                 */
/*                                                                           */
/* Copyright (C) 2001 Magnus Damm <damm@opensource.se>                       */
/*                                                                           */
/* Parts of the code based on:                                               */
/*                                                                           */
/* atftp:  Copyright (C) 2000 Jean-Pierre Lefebvre <helix@step.polymtl.ca>   */
/*                        and Remi Lefebvre <remi@debian.org>                */
/*                                                                           */
/* utftp:  Copyright (C) 1999 Uwe Ohse <uwe@ohse.de>                         */
/*                                                                           */
/* This program is free software; you can redistribute it and/or modify      */
/* it under the terms of the GNU General Public License as published by      */
/* the Free Software Foundation; either version 2 of the License, or         */
/* (at your option) any later version.                                       */

/*                                                                           */
/* This program is distributed in the hope that it will be useful,           */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU          */
/* General Public License for more details.                                  */
/*                                                                           */
/* You should have received a copy of the GNU General Public License         */
/* along with this program; if not, write to the Free Software               */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   */
/*                                                                           */
/* ------------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <fcntl.h>

#include "busybox.h"

#include <sys/ioctl.h>

#include <linux/fs.h>
#include <linux/mtd/mtd.h>
//#define BB_FEATURE_TFTP_DEBUG

#define UPGRADE_ONLY

#ifdef UPGRADE_ONLY

#define	UP_DEBUG
#define	SIZE_PER_SEC		512

#define FLASHDEV 				("/dev/mtd_block2")
#define START_ADDR				(0x20000)
#define CODE_IMAGE_OFFSET		(0x20000) //neo@

// IC1500(Wg) definition
#define IC1500WG_FW_HEADER		((char *)"IC1500WG")
#define IC1500WG_IMAGE_OFFSET	(0x20000)
#define IC1500_FW_HEADER		((char *)"IC1500  ")
#define IC1500L_FW_HEADER		((char *)"IC1500L ")
#define IC1500S_FW_HEADER		((char *)"IC1500S ")
#define IC1500_IMAGE_OFFSET		(0x20000)
//#define IC1500_FW_HEADER_FS		((char *)"IC1500FS")
//#define IC1500_IMAGE_OFFSET_FS	(0xB0000)


#define TFTP_FILE "/etc/tftp_file"
#define HEADER_LEN	(8)
int	_flashfd;
int 	_tftpfd;				//J01
unsigned long 	filebuff=0xffffffff;

#endif

/* we don't need #ifdefs with these constants and optimization... */

//#ifdef BB_FEATURE_TFTP_GET
//#define BB_TFTP_GET (1 << 0)
//#else
//#define BB_TFTP_GET 0
//#endif

//#ifdef BB_FEATURE_TFTP_PUT
//#define BB_TFTP_PUT (1 << 1)
//#else
//#define BB_TFTP_PUT 0
//#endif

#define BB_TFTP_GET 1
#define BB_TFTP_PUT 0

#ifdef BB_FEATURE_TFTP_DEBUG
#define BB_TFTP_DEBUG 1
#else
#define BB_TFTP_DEBUG 0
#endif

#define BB_TFTP_NO_RETRIES 5
#define BB_TFTP_TIMEOUT    5	/* seconds */

#define	RRQ	1			/* read request */
#define	WRQ	2			/* write request */
#define	DATA	3		/* data packet */
#define	ACK	4			/* acknowledgement */
#define	ERROR	5		/* error code */

#define BUFSIZE (512+4)

static const char *tftp_error_msg[] = {
	"Undefined error",
	"File not found",
	"Access violation",
	"Disk full or allocation error",
	"Illegal TFTP operation",
	"Unknown transfer ID",
	"File already exists",
	"No such user"
};

static inline int tftp(int cmd, struct hostent *host,
					   char *serverfile, int localfd, int port)
{
	struct sockaddr_in sa;
	int socketfd;
	struct timeval tv;
	fd_set rfds;
	struct sockaddr_in from;
	socklen_t fromlen;
	char *cp;
	unsigned short tmp;
	int len, opcode, finished;
	int timeout, block_nr;
	unsigned long counter=0;
	
	unsigned char webHEADER[HEADER_LEN];	// Rex

	int skipLen = HEADER_LEN;

#ifdef UPGRADE_ONLY
	int	need_erase;
#endif

	//RESERVE_BB_BUFFER(buf, BUFSIZE);

    char *buf=xmalloc(BUFSIZE + 4);
	
#ifdef UPGRADE_ONLY

	if((_flashfd = open(FLASHDEV, O_RDWR))<0)
	{
		printf("\ncan not open flash device\n");
		perror("\n");
		exit(1);
	}

							
	need_erase = 1;
	
#endif

	opcode = finished = timeout = 0;
	block_nr = 1;

	if ((socketfd = socket(PF_INET, SOCK_DGRAM, 0)) < 0) {
		printf("socket");
		return EXIT_FAILURE;
	}

	len = sizeof(sa);

	memset(&sa, 0, len);
	bind(socketfd, (struct sockaddr *)&sa, len);

	sa.sin_family = host->h_addrtype;
	sa.sin_port = htons(port);
	memcpy(&sa.sin_addr, (struct in_addr *) host->h_addr,
		   sizeof(sa.sin_addr));

	/* build opcode */

	if (cmd & BB_TFTP_GET) {
		opcode = RRQ;
	}

	if (cmd & BB_TFTP_PUT) {
		opcode = WRQ;
	}

	while (1) {


		/* build packet of type "opcode" */

		cp = buf;

		*((unsigned short *) cp) = htons(opcode);

		cp += 2;

		/* add filename and mode */

		if ((BB_TFTP_GET && (opcode == RRQ)) ||
			(BB_TFTP_PUT && (opcode == WRQ))) {

			while (cp != &buf[BUFSIZE - 1]) {
				if ((*cp = *serverfile++) == '\0')
					break;
				cp++;
			}

			if ((*cp != '\0') || (&buf[BUFSIZE - 1] - cp) < 7) {
				printf("too long server-filename");
				break;
			}

			memcpy(cp + 1, "octet", 6);
			cp += 7;
		}

		/* add ack and data */

		if ((BB_TFTP_GET && (opcode == ACK)) ||
			(BB_TFTP_PUT && (opcode == DATA))) {

			*((unsigned short *) cp) = htons(block_nr);

			cp += 2;

			block_nr++;

			if (BB_TFTP_PUT && (opcode == DATA)) {
				
				len = read(_flashfd, cp, BUFSIZE - 4);

				if (len < 0) {
					printf("read");
					break;
				}

				if (len != (BUFSIZE - 4)) {
					finished++;
				}

				cp += len;
				
			} else if (finished) {
				printf("\n1 finished break");
				//break;   //Jason(2003/08/05)
			}
		}


		/* send packet */


		do {

			len = cp - buf;

			if (BB_TFTP_DEBUG) {
				printf("\nsending %u bytes", len);

				for (cp = buf; cp < &buf[len]; cp++)
					printf("%02x ", *cp);
			}				

			if (sendto(socketfd, buf, len, 0,
					   (struct sockaddr *) &sa, sizeof(sa)) < 0) {
				printf("send");
				len = -1;
				break;
			}


			if (finished) {
				printf("\n2 finished break");
				break;
			}
			
			/* receive packet */
			
			memset(&from, 0, sizeof(from));
			fromlen = sizeof(from);

			tv.tv_sec = BB_TFTP_TIMEOUT;
			tv.tv_usec = 0;

			FD_ZERO(&rfds);
			FD_SET(socketfd, &rfds);

			switch (select(FD_SETSIZE, &rfds, NULL, NULL, &tv)) {
			case 1:
				len = recvfrom(socketfd, buf,
							   BUFSIZE, 0,
							   (struct sockaddr *) &from, &fromlen);

				if (len < 0) {
					printf("case 1 (len<0)");
					break;
				}

				timeout = 0;

				if (sa.sin_port == htons(port)) {
					sa.sin_port = from.sin_port;
					break;
				}

				if (sa.sin_port == from.sin_port) {
					break;
				}

				/* fall-through for bad packets! */
				/* discard the packet - treat as timeout */

			case 0:
				printf("\ntimeout ");
				if (!timeout) {
					timeout = BB_TFTP_NO_RETRIES;
				} 
				else {
					timeout--;
				}

				if (!timeout) {
					len = -1;
					printf("last timeout");
				}
				break;

			default:
				printf("select");
				len = -1;
			}

		} while (timeout && (len >= 0));

		if (len < 0) {
			printf("\nlen < 0");
			break;
		}

		//===> Break immediately after the last ack is send
		if (finished) {
			printf("\n3 finished break");
			break;
		}//<<=== Jason(2003/08/05)
		
		/* process received packet */


		opcode = ntohs(*((unsigned short *) buf));
		tmp = ntohs(*((unsigned short *) &buf[2]));
	
		if (BB_TFTP_DEBUG) {
			printf("\nreceived %d bytes: %04x %04x", len, opcode, tmp);
		}

		if (BB_TFTP_GET && (opcode == DATA)) {

			if (tmp == block_nr) {
#ifdef UPGRADE_ONLY
				if (need_erase) {
					if((_tftpfd = open(TFTP_FILE, O_RDONLY)) == -1) {
						printf("header: [%c%c%c%c%c%c%c%c]\n", buf[4], buf[5], buf[6], buf[7], buf[8], buf[9], buf[10], buf[11]);
						if(!memcmp(&buf[4], IC1500WG_FW_HEADER, HEADER_LEN)) {
							printf("seek to 0x%X\n", IC1500WG_IMAGE_OFFSET);
							lseek(_flashfd, IC1500WG_IMAGE_OFFSET, SEEK_SET);
						}
						else if (!memcmp(&buf[4], IC1500_FW_HEADER, HEADER_LEN) || 
								 !memcmp(&buf[4], IC1500L_FW_HEADER, HEADER_LEN) ||
								 !memcmp(&buf[4], IC1500S_FW_HEADER, HEADER_LEN)) {
							printf("seek to 0x%X\n", IC1500_IMAGE_OFFSET);
							lseek(_flashfd, IC1500_IMAGE_OFFSET, SEEK_SET);
						}
						/*
						else if (!memcmp(&buf[4], IC1500_FW_HEADER_FS, HEADER_LEN)) {
							printf("seek to 0x%X\n", IC1500_IMAGE_OFFSET_FS);
							lseek(_flashfd, IC1500_IMAGE_OFFSET_FS, SEEK_SET);
						}
						*/
						else {
							printf("header was wrong!\n");
							exit(1);
						}
					}
					else
						close(_tftpfd);
					need_erase = 0;
				}
				counter++;
#if 1
				if (skipLen > 0) {
					//neo@: IC1500 need skip 8 bytes header
					len = write(_flashfd, (&buf[4]) + skipLen, len - 4 - skipLen);
					len += skipLen; //in order to pass check below : if (len != (BUFSIZE - 4)
					skipLen = 0;
				}
				else
					len = write(_flashfd, &buf[4], len - 4);
#else
				len =len - 4;
#endif
				if(counter == 512)
				{	
					ioctl(_flashfd,BLKFLSBUF,0);
					counter=0;
				}	
				//printf(".");
#else
				len = write(localfd, &buf[4], len - 4);
#endif

				if (len < 0) {
					printf("\nwrite error");
					break;
				}

				if (len != (BUFSIZE - 4)) {
					printf("\nlen != BUFSIZE-4  len:%d\n", len);
					finished++;
				}

				opcode = ACK;

				continue;
			}
		}

		if (BB_TFTP_PUT && (opcode == ACK)) {

			if (tmp == (block_nr - 1)) {
				if (finished) {
					break;
				}

				opcode = DATA;
				continue;
			}
		}

		if (opcode == ERROR) {
			char *msg = NULL;

			if (buf[4] != '\0') {
				msg = &buf[4];
				buf[BUFSIZE - 1] = '\0';
			} else if (tmp < (sizeof(tftp_error_msg) / sizeof(char *))) {
				msg = (char *) tftp_error_msg[tmp];
			}

			if (msg) {
				printf("server says: %s", msg);
			}

			break;
		}
	}

#ifdef UPGRADE_ONLY
	close(_flashfd);
/*
{
	unsigned short checksum, checksum_org;
	unsigned long readLen = 0;
	unsigned char readBuf[1024];
	unsigned long i;
	
	if((_flashfd = open(FLASHDEV, O_RDWR))<0)
	{
		printf("\ncan not open flash device ");
		exit(1);
	}

	for(i=0; i < 0x10000; i+=0x1000)
	{
		lseek(_flashfd, START_ADDR + i, SEEK_SET);
		read(_flashfd, readBuf, 12);
		if(memcmp(readBuf, "CSYS", 4))
			continue;
		readLen = *((unsigned long *)&readBuf[8]);
		printf("\n\nFound CSYS image in %lx, Len = %lx\n", START_ADDR + i, readLen);
		break;
	}

	if ( i == 0x10000)
	{    
		close(_flashfd);
		printf("\n\nUpgrade Failed\n");
		return EXIT_FAILURE;
	}

	checksum = 0;
	comment by neo@ : This is unnecessary.

	for(i=0 ;i < (readLen); i += 2)
	{
		read(_flashfd, readBuf, 2);
		checksum += *(unsigned short*)readBuf;
	}
	
	checksum_org = *(unsigned short*)readBuf;
	
	printf("\n checksum_org = %x, checksum=%x\n", checksum_org, checksum);
	

	close(_flashfd);

	if(checksum != 0)
		return EXIT_FAILURE;
}
*/	
#endif
	close(socketfd);
	return finished ? EXIT_SUCCESS : EXIT_FAILURE;
}

int tftp_main(int argc, char **argv)
{
	char *cp, *s;
	char *serverstr;
	struct hostent *host;
	char *serverfile;
	char *localfile;
	int cmd, flags, fd, bad;
#if 1	/* SJC */
	u_short	port=69;
#endif

	host = (void *) serverstr = serverfile = localfile = NULL;
	flags = cmd = 0;
	bad = 1;

#ifdef UPGRADE_ONLY
	if (argc > 2) {
		if (BB_TFTP_GET && (strcmp(argv[1], "get") == 0)) {
			cmd = BB_TFTP_GET;
			flags = O_WRONLY | O_CREAT;
			serverstr = argv[2];
			localfile = NULL;
		}
		port = 69;
		if (argc > 3) {
			port = atoi(argv[3]);
			if (port == 0)
				port = 69;
		}
	}
	else{
		printf ("tftp get xxx.xxx.xxx.xxx:filename (port)\n");
		exit (0);
	}
#else
	if (argc > 3) {
		if (BB_TFTP_GET && (strcmp(argv[1], "get") == 0)) {
			cmd = BB_TFTP_GET;
			flags = O_WRONLY | O_CREAT;
			serverstr = argv[2];
			localfile = argv[3];
		}

		if (BB_TFTP_PUT && (strcmp(argv[1], "put") == 0)) {
			cmd = BB_TFTP_PUT;
			flags = O_RDONLY;
			localfile = argv[2];
			serverstr = argv[3];
		}

	}
#endif

	if (!(cmd & (BB_TFTP_GET | BB_TFTP_PUT))) {
;//		show_usage();
	}

	for (cp = serverstr; *cp != '\0'; cp++)
		if (*cp == ':')
			break;

	if (*cp == ':') {

		serverfile = cp + 1;

//		s = xstrdup(serverstr);
		s = strdup(serverstr);
		s[cp - serverstr] = '\0';

		host = xgethostbyname(s);

		free(s);
	}

	if (BB_TFTP_DEBUG) {
		printf("using server \"%s\", serverfile \"%s\","
			   "localfile \"%s\".\n",
			   inet_ntoa(*((struct in_addr *) host->h_addr)),
			   serverfile, localfile);
	}

#ifdef UPGRADE_ONLY
	fd = -1;
#else
	if ((fd = open(localfile, flags, 0644)) < 0) {
		perror_msg_and_die("local file");
	}
#endif

#if 1	/* SJC */
	flags = tftp(cmd, host, serverfile, fd, port);
#else
	flags = tftp(cmd, host, serverfile, fd, 69);
	close(fd);
#endif
	printf("\ntftp ok ! ");
//	system("reboot");	

	return flags;
}
